#! /usr/bin/env python
# -*- coding: utf-8 -*-

##################################################
# ロボットが棚の前まで移動し、左手で商品を把持する
# 左手で商品を把持したまま人の前まで移動する
# 把持した商品を人に渡したら、初期位置まで移動する
##################################################

import rospy
import math
import copy
import moveit_commander
import moveit_msgs.msg
from geometry_msgs.msg import Quaternion, Pose, PoseStamped, Vector3
from tf.transformations import quaternion_from_euler
import actionlib
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal

######################################
# Global Navigation
######################################

waypoints_1 = [
    [(0.0, 0.0, 0.0),(0.0, 0.0, 0.0, 1.0)],
    [(-2.903, 3.404, 0.075), (0.0, 0.0, 0.700, 0.700)],
    [(-2.418, 6.850, 0.075), (0.0, 0.0, 0.023, 1)]
]

waypoints_2 = [
    [(-0.472, 4.237, 0.075),(0.0, 0.0, -0.703, 0.711)],
    [(-2.847, 2.995, 0.075),(0.0, 0.0, 0.703, 0.711)]
]

waypoints_3 = [
    [(0.0, 0.0, 0.0),(0.0, 0.0, 0.0, 1.0)]
]

def goal_pose(pose):
    goal_pose = MoveBaseGoal()
    goal_pose.target_pose.header.frame_id = 'map'
    goal_pose.target_pose.pose.position.x = pose[0][0]
    goal_pose.target_pose.pose.position.y = pose[0][1]
    goal_pose.target_pose.pose.position.z = pose[0][2]
    goal_pose.target_pose.pose.orientation.x = pose[1][0]
    goal_pose.target_pose.pose.orientation.y = pose[1][1]
    goal_pose.target_pose.pose.orientation.z = pose[1][2]
    goal_pose.target_pose.pose.orientation.w = pose[1][3]
    return goal_pose

SLEEP_TIME = 2.0

def euler_to_quaternion(role, pitch, yaw):
    q = quaternion_from_euler(role, pitch, yaw)
    return Quaternion(q[0], q[1], q[2], q[3])

######################################
# MAIN
######################################

def main():
    
    # init node
    rospy.init_node("final_task_node")
    
    # actionlib client declaration 
    client = actionlib.SimpleActionClient('move_base', MoveBaseAction)
    client.wait_for_server()

    # configuration for moveit
    robot = moveit_commander.RobotCommander()
    scene = moveit_commander.PlanningSceneInterface()
    
    # group name
    rarm = moveit_commander.MoveGroupCommander("rarm")
    larm = moveit_commander.MoveGroupCommander("larm")
    lifter = moveit_commander.MoveGroupCommander("lifter")
    upper = moveit_commander.MoveGroupCommander("upper_body")
    waist = moveit_commander.MoveGroupCommander("waist")
    rarm_waist = moveit_commander.MoveGroupCommander("rarm_with_waist")
    larm_waist = moveit_commander.MoveGroupCommander("larm_with_waist")
    rarm_waist.set_end_effector_link("r_eef_pick_link")
    larm_waist.set_end_effector_link("l_eef_pick_link")

    # reset pose
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    # waypoints_1
    # 棚の前まで移動する
    for pose in waypoints_1:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(1.5)

    ###########################################
    # pick a merchandise（商品棚から商品を取る）
    ###########################################

    # trun left
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[0] = 1.57
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True) 

    rospy.sleep(1.5)

    print ("")
    print ("Pose 1: approach")
    pose_goal1 = PoseStamped()
    pose_goal1.header.frame_id = "base_link"
    pose_goal1.pose.orientation.x = 0.541
    pose_goal1.pose.orientation.y = 0.439
    pose_goal1.pose.orientation.z = 0.471
    pose_goal1.pose.orientation.w = 0.541
    pose_goal1.pose.position.x = -0.140
    pose_goal1.pose.position.y = 0.517
    pose_goal1.pose.position.z = 1.180
    larm_waist.set_pose_target(pose_goal1)
    print ("LARM_WITH_WAIST", larm_waist.go(wait=True))

    rospy.sleep(3)

    print ("")
    print ("Pose 2: hand down")
    pose_goal2 = PoseStamped()
    pose_goal2.header.frame_id = "base_link"
    pose_goal2.pose.orientation.x = 0.541
    pose_goal2.pose.orientation.y = 0.439
    pose_goal2.pose.orientation.z = 0.471
    pose_goal2.pose.orientation.w = 0.541
    pose_goal2.pose.position.x = -0.140
    pose_goal2.pose.position.y = 0.517
    pose_goal2.pose.position.z = 1.150
    larm_waist.set_pose_target(pose_goal2)
    print ("LARM_WITH_WAIST", larm_waist.go(wait=True))

    rospy.sleep(1)

    # left hand grasp
    ##########################################
    # print ("")
    # print ("left hand grasp")
    # response = service(1, 'grasp', 100)

    # rospy.sleep(5)
    ##########################################

    print ("")
    print ("pose 3: hand up")
    pose_goal3 = PoseStamped()
    pose_goal3.header.frame_id = "base_link"
    pose_goal3.pose.orientation.x = 0.541
    pose_goal3.pose.orientation.y = 0.439
    pose_goal3.pose.orientation.z = 0.471
    pose_goal3.pose.orientation.w = 0.541
    pose_goal3.pose.position.x = -0.140
    pose_goal3.pose.position.y = 0.517
    pose_goal3.pose.position.z = 1.253
    larm_waist.set_pose_target(pose_goal3)
    print ("LARM_WITH_WAIST", larm_waist.go(wait=True))

    rospy.sleep(1)

    # reset pose
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    # waypoints_2
    # 人の前まで移動する
    for pose in waypoints_2:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(5)

    ######################################
    # take a merchandise（商品を渡す）
    ######################################

    # trun left
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[0] = 1.57
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True) 

    rospy.sleep(1.5)

    print ("")
    print ("Pose 4: approach")
    pose_goal4 = PoseStamped()
    pose_goal4.header.frame_id = "base_link"
    pose_goal4.pose.orientation.x = 0.687
    pose_goal4.pose.orientation.y = 0.182
    pose_goal4.pose.orientation.z = -0.202
    pose_goal4.pose.orientation.w = 0.674
    pose_goal4.pose.position.x = 0.554
    pose_goal4.pose.position.y = 0.187
    pose_goal4.pose.position.z = 1.176
    larm.set_pose_target(pose_goal4)
    print ("LARM_WITH_WAIST", larm_waist.go(wait=True))

    rospy.sleep(3)

    # left hand release
    ##############################################
    # print ("")
    # print ("left hand release")
    # response = service(1, 'release', 100)

    # rospy.sleep(3)
    ##############################################

    # reset pose
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    # waypoints_3
    # 初期位置まで移動する
    for pose in waypoints_3:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(5)

    print ("")
    print ("task finished!")
    
if __name__ == '__main__':
    try:
        main()
    except rospy.ROSInterruptException:
        pass